#include "libvdev/sglib.h"
#include "libvdev/util.h"
#include "libvdev/misc.h"
#include "libvdev/sbuf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdbool.h>
#include <mntent.h>
#include <sys/stat.h>

const char *progname = "vdev_bind_input_devices";
const char *mtab = "/etc/mtab";

typedef char* cstr;

// prototypes
SGLIB_DEFINE_VECTOR_PROTOTYPES (cstr);
SGLIB_DEFINE_VECTOR_FUNCTIONS (cstr);

// free a list of cstr vectors
// always succeeds
static int vdev_cstr_vector_free_all(struct sglib_cstr_vector* vec)
{
        // free all strings
        for (int i = 0; i < sglib_cstr_vector_size(vec); i++) {
                if (sglib_cstr_vector_at(vec, i) != NULL) {
                        free( sglib_cstr_vector_at(vec, i));
                        sglib_cstr_vector_set(vec, NULL, i);
                }
        }
        return 0;
}

// 'buf' must be freed after usage
static void get_sysfs_path(char **buf)
{
        struct mntent *e;
        FILE* fstab = NULL;
   
        fstab = setmntent(mtab, "r");
        if (!fstab) {
                vdev_error("%s: setmntent(): error trying to open /etc/mtab: '%s'\n", progname, strerror (errno));
                exit(EXIT_FAILURE);
        }
    
        *buf = (char*)malloc(sizeof(char) * 32);
        if (!*buf) {
                vdev_error ("%s: Memory allocation failure: '%s'\n", progname, strerror (errno));
                exit(EXIT_FAILURE);
        }
   
        *buf[0] = '\0';
         
        while ((e = getmntent(fstab))) {
                if (!strcmp(e->mnt_type, "sysfs")) {
                        sprintf( *buf, "%s", e->mnt_dir );
                        break;
                }
        }
   
        endmntent(fstab);
}

int main (int argc, char **argv)
{
        int rc;
        struct sglib_cstr_vector aux;
        char sysfs_mountpoint[PATH_MAX+1];
        char line[PATH_MAX+1];
        char *buf = NULL;
        FILE *fp = NULL;
        const char *file = "/proc/bus/input/devices";
        
        send_to_background();
   
        if (access(mtab, F_OK) != 0) {
                snprintf(sysfs_mountpoint, PATH_MAX, "/sys"); 
        } else {
                get_sysfs_path(&buf);
                if (buf && buf[0] != '\0') {
                        snprintf(sysfs_mountpoint, PATH_MAX, buf);       
                        free(buf);   
                } else {
                        vdev_error("%s: Cannot get sysfs path\n", progname);
                        exit(EXIT_FAILURE);
                }
        }
                
        fp = fopen(file, "r");
        if (!fp) {
                vdev_error("fopen(%s): %s\n", file, strerror(errno));
                exit(EXIT_FAILURE);
        }
        sglib_cstr_vector_init(&aux);
        while (fgets(line, PATH_MAX+1, fp)) {
                if (!strncmp(line, "S: Sysfs=", strlen("S: Sysfs="))) {
                        sbuf_t buf;
                        sbuf_t item;
                        sbuf_t driver;
                        char linkpath[PATH_MAX+1] = {0};
                        size_t linkpath_len = PATH_MAX;
                        char *path = NULL;
                        char *address = NULL;
                        line[strcspn(line, "\n")] = '\0';
                        char *tmp = rindex(line, '=')+1;
                        
                        sbuf_init(&item);
                        sbuf_concat(&item, 3, sysfs_mountpoint, tmp, "/device");
                        if (access(item.buf, F_OK) != 0) {
                                sbuf_free(&item);
                                continue;
                        }
                        rc = readlink(item.buf, linkpath, linkpath_len);
                        if (rc < 0) {
                                vdev_error("%s: readlink('%s') %s\n", progname, linkpath, strerror(errno));
                                sbuf_free(&item);
                                vdev_cstr_vector_free_all(&aux);       
                                sglib_cstr_vector_free(&aux);
                                fclose(fp);
                                exit(EXIT_FAILURE);
                        }
                        sbuf_free(&item);
                        
                        sbuf_init(&buf);
                        sbuf_concat(&buf, 4, sysfs_mountpoint, tmp, "/", linkpath);
                        path = realpath(buf.buf, NULL);
                        sbuf_free(&buf);

                        address = vdev_dirname(path, NULL);
                        if (address == NULL) {
                                free(path);
                                vdev_cstr_vector_free_all(&aux);       
                                sglib_cstr_vector_free(&aux);
                                fclose(fp);
                                exit(EXIT_FAILURE);
                        }
                        free(path);

                        sbuf_init(&driver);
                        sbuf_concat(&driver, 2, address, "/driver");
                        if (access(driver.buf, F_OK) == 0) {
                                sbuf_t cmd;
                                char delim[2] = "/";
                                char* token = NULL;
                                char *copy = NULL;
                                char buspath[PATH_MAX+1] = {0};
                                size_t buspath_len = PATH_MAX;
                                rc = readlink(driver.buf, buspath, buspath_len);
                                if (rc < 0) {
                                        vdev_error("%s: readlink('%s') %s\n", progname, buspath, strerror(errno));
                                        free(address);
                                        sbuf_free(&driver);
                                        vdev_cstr_vector_free_all(&aux);       
                                        sglib_cstr_vector_free(&aux);
                                        fclose(fp);
                                        exit(EXIT_FAILURE);
                                }
                                sbuf_init(&cmd);
                                sbuf_concat(&cmd, 3, "/bin/echo \"", rindex(address, '/')+1, "\" | sudo tee /sys");
                                token = strtok(buspath, delim);
                                while (token != NULL) {
                                        if (strcmp(token, "..")) {
                                                sbuf_addstr(&cmd, "/");
                                                sbuf_addstr(&cmd, token);
                                        }
                                        token = strtok(NULL, delim);
                                }
                                copy = vdev_strdup_or_null(cmd.buf);
                                if (!copy) {
                                        free(address);
                                        sbuf_free(&cmd);
                                        sbuf_free(&driver);
                                        vdev_cstr_vector_free_all(&aux);       
                                        sglib_cstr_vector_free(&aux);
                                        fclose(fp);
                                        return -ENOMEM;
                                }                                
                                rc = sglib_cstr_vector_push_back(&aux, copy);
                                if (rc < 0) {
                                        vdev_error ("%s: sglib_cstr_vector_push_back('%s'): '%s'\n", progname, copy, strerror(errno));
                                        free(copy);
                                        free(address);
                                        sbuf_free(&cmd);
                                        sbuf_free(&driver);
                                        vdev_cstr_vector_free_all(&aux);       
                                        sglib_cstr_vector_free(&aux);
                                        fclose(fp);
                                        return rc;
                                }
                                sbuf_free(&cmd);
                        }
                        free(address);
                        sbuf_free(&driver);
                }
        }
        fclose(fp);
                
        for (int i = 0; i < sglib_cstr_vector_size(&aux); i++) {
                FILE *pfin = NULL;
                pid_t pid;
                int wstatus;
                sbuf_t unbind_cmd;
                sbuf_init(&unbind_cmd);
                sbuf_concat(&unbind_cmd, 2, sglib_cstr_vector_at(&aux, i), "/unbind >/dev/null");
                pfin = epopen(unbind_cmd.buf, &pid);
                if (pfin) { 
                        fclose(pfin);
                        waitpid(pid, &wstatus, 0);
                }
                sbuf_free(&unbind_cmd);
        }
        
        for (int i = 0; i < sglib_cstr_vector_size(&aux); i++) {
                FILE *pfin = NULL;
                pid_t pid;
                int wstatus;
                sbuf_t bind_cmd;
                sbuf_init(&bind_cmd);
                sbuf_concat(&bind_cmd, 2, sglib_cstr_vector_at(&aux, i), "/bind >/dev/null");
                pfin = epopen(bind_cmd.buf, &pid);
                if (pfin) { 
                        fclose(pfin);
                        waitpid(pid, &wstatus, 0);
                }
                sbuf_free(&bind_cmd);
        }

        vdev_cstr_vector_free_all(&aux);       
        sglib_cstr_vector_free(&aux);
                
        return 0;
}
